import logging
import random
from typing import List, Literal, TypedDict

from random_slugs.words import get_words_by_category

logger = logging.getLogger(__name__)

FORMAT_OPTION = Literal["kebab", "snake", "camel"]

DEFAULT_NUMBER_OF_WORDS = 3
DEFAULT_FORMAT_OPTION = "kebab"

Options = TypedDict(
    "Options",
    {"parts_of_speech": List[str], "categories": dict, "format": FORMAT_OPTION, "seed": None},
)


class RandomSlugsError(Exception):
    """
    Custom exception for all errors related to slug generation.
    """

    pass


class RandomSlugConfigError(RandomSlugsError):
    """
    Exception for all configuration errors related.
    """

    pass


def generate_slug(num_of_words: int = DEFAULT_NUMBER_OF_WORDS, options: Options = None) -> str:
    """
    Generates a random slug based on the given options.
    """
    words = []
    opts = _get_default_options(num_of_words)

    if options is not None:
        _validate_options(options)
        opts.update(options)

    if isinstance(opts["seed"], (int, str, float, bytes, bytearray)):
        random.seed(opts["seed"])

    for part in opts["parts_of_speech"]:
        if part in opts["categories"]:
            categories = opts["categories"][part]
        else:
            categories = []
        candidates = get_words_by_category(part, categories=categories)
        candidate = random.choice(candidates)
        words.append(candidate)

    return format_slug(words, opts["format"])


def _get_default_parts_of_speech(num_of_words: int) -> List[str]:
    parts_of_speech = []
    for _ in range(num_of_words - 1):
        parts_of_speech.append("adjectives")
    parts_of_speech.append("nouns")
    return parts_of_speech


def _get_default_options(num_of_words: int = DEFAULT_NUMBER_OF_WORDS) -> Options:
    default_options = {
        "parts_of_speech": _get_default_parts_of_speech(num_of_words),
        "categories": {},
        "format": DEFAULT_FORMAT_OPTION,
        "seed": None,
    }
    return default_options


def _validate_options(options: Options) -> None:
    if "parts_of_speech" in options:
        for part in options["parts_of_speech"]:
            if part not in ["adjectives", "nouns"]:
                raise RandomSlugConfigError(f"Invalid part of speech: {part}")
    if "categories" in options:
        for part, _categories in options["categories"].items():
            if part not in ["adjectives", "nouns"]:
                raise RandomSlugConfigError(f"Invalid part of speech: {part}")
    if "format" in options:
        if options["format"] not in ["kebab", "snake", "camel"]:
            raise RandomSlugConfigError(f"Invalid format: {options['format']}")
    if "seed" in options:
        if not isinstance(options["seed"], (int, str, float, bytes, bytearray)):
            raise RandomSlugConfigError(f"Invalid seed: {options['seed']}")


def format_slug(words: List[str], fmt: FORMAT_OPTION = DEFAULT_FORMAT_OPTION) -> str:
    """
    Formats a list of words into a slug based on the given format.
    """
    if fmt == "kebab":
        return "-".join(words)
    elif fmt == "snake":
        return "_".join(words)
    elif fmt == "camel":
        return "".join([words[0]] + [word.capitalize() for word in words[1:]])


def get_total_unique_slugs(
    number_of_words: int = DEFAULT_NUMBER_OF_WORDS, options: Options = None
) -> int:
    """
    Returns the total number of unique slugs that can be generated with the given options.
    """
    if options is None:
        options = _get_default_options(number_of_words)

    num_adjectives = len(
        get_words_by_category("adjectives", options.get("categories", {}).get("adjectives", []))
    )
    num_nouns = len(get_words_by_category("nouns", options.get("categories", {}).get("nouns", [])))
    nums = {"adjectives": num_adjectives, "nouns": num_nouns}

    num_words = number_of_words or DEFAULT_NUMBER_OF_WORDS
    parts_of_speech = options.get("parts_of_speech", _get_default_parts_of_speech(num_words))

    combos = 1
    for i in range(num_words):
        combos *= nums[parts_of_speech[i]]

    return combos



from typing import List, Literal, Optional, Tuple, TypedDict

Category = Literal[
    "animal",
    "color",
    "transport",
    "pokemon",
    "technology",
    "profession",
    "thing",
    "quantity",
    "size",
    "emotion",
    "condition",
]

WordEntry = Tuple[str, List[Category]]


class Vocabulary(TypedDict):
    nouns: List[WordEntry]
    adjectives: List[WordEntry]


vocabulary_kwargs = {
    "nouns": [
        ("tiger", ["animal", "color"]),
        ("elephant", ["animal", "color"]),
        ("camel", ["animal"]),
        ("giraffe", ["animal"]),
        ("elephant", ["animal"]),
        ("otter", ["animal"]),
        ("mongoose", ["animal"]),
        ("porcupine", ["animal"]),
        ("jaguar", ["animal"]),
        ("gorilla", ["animal"]),
        ("koala", ["animal"]),
        ("meerkat", ["animal"]),
        ("ocelot", ["animal"]),
        ("buffalo", ["animal"]),
        ("platypus", ["animal"]),
        ("anteater", ["animal"]),
        ("armadillo", ["animal"]),
        ("cheetah", ["animal"]),
        ("chameleon", ["animal"]),
        ("cat", ["animal"]),
        ("lion", ["animal"]),
        ("walrus", ["animal"]),
        ("quokka", ["animal"]),
        ("lynx", ["animal"]),
        ("sloth", ["animal"]),
        ("wolverine", ["animal"]),
        ("lemur", ["animal"]),
        ("carvan", ["transport"]),
        ("boat", ["transport"]),
        ("plane", ["transport"]),
        ("train", ["transport"]),
        ("bicycle", ["transport"]),
        ("motorcycle", ["transport"]),
        ("scooter", ["transport"]),
        ("skateboard", ["transport"]),
        ("car", ["transport"]),
        ("truck", ["transport"]),
        ("bus", ["transport"]),
        ("subway", ["transport"]),
        ("helicopter", ["transport"]),
        ("rocket", ["transport"]),
        ("spaceship", ["transport"]),
        ("tractor", ["transport"]),
        ("golfcart", ["transport"]),
        ("segway", ["transport"]),
        ("pikachu", ["pokemon"]),
        ("charmander", ["pokemon"]),
        ("bulbasaur", ["pokemon"]),
        ("squirtle", ["pokemon"]),
        ("snorlax", ["pokemon"]),
        ("mewtwo", ["pokemon"]),
        ("mew", ["pokemon"]),
        ("articuno", ["pokemon"]),
        ("zapdos", ["pokemon"]),
        ("moltres", ["pokemon"]),
        ("raikou", ["pokemon"]),
        ("entei", ["pokemon"]),
        ("suicune", ["pokemon"]),
        ("lugia", ["pokemon"]),
        ("celebi", ["pokemon"]),
        ("rayquaza", ["pokemon"]),
        ("jirachi", ["pokemon"]),
        ("deoxys", ["pokemon"]),
        ("uxie", ["pokemon"]),
        ("mesprit", ["pokemon"]),
        ("azelf", ["pokemon"]),
        ("dialga", ["pokemon"]),
        ("palkia", ["pokemon"]),
        ("heatran", ["pokemon"]),
        ("giratina", ["pokemon"]),
        ("cpu", ["technology"]),
        ("gpu", ["technology"]),
        ("ram", ["technology"]),
        ("rom", ["technology"]),
        ("ssd", ["technology"]),
        ("hdd", ["technology"]),
        ("usb", ["technology"]),
        ("hdmi", ["technology"]),
        ("vga", ["technology"]),
        ("dvi", ["technology"]),
        ("ethernet", ["technology"]),
        ("bluetooth", ["technology"]),
        ("floppy", ["technology"]),
        ("zip", ["technology"]),
        ("vhs", ["technology"]),
        ("cassette", ["technology"]),
        ("vinyl", ["technology"]),
        ("floppy", ["technology"]),
        ("laserdisc", ["technology"]),
        ("scientist", ["profession"]),
        ("doctor", ["profession"]),
        ("engineer", ["profession"]),
        ("programmer", ["profession"]),
        ("teacher", ["profession"]),
        ("professor", ["profession"]),
        ("student", ["profession"]),
        ("nurse", ["profession"]),
        ("artist", ["profession"]),
        ("musician", ["profession"]),
        ("actor", ["profession"]),
        ("actress", ["profession"]),
        ("singer", ["profession"]),
        ("dancer", ["profession"]),
        ("chef", ["profession"]),
        ("firefighter", ["profession"]),
        ("police", ["profession"]),
        ("officer", ["profession"]),
        ("soldier", ["profession"]),
        ("pilot", ["profession"]),
        ("astronaut", ["profession"]),
        ("lawyer", ["profession"]),
        ("judge", ["profession"]),
        ("politician", ["profession"]),
        ("president", ["profession"]),
        ("king", ["profession"]),
        ("queen", ["profession"]),
        ("prince", ["profession"]),
        ("princess", ["profession"]),
        ("emperor", ["profession"]),
        ("empress", ["profession"]),
        ("dictator", ["profession"]),
        ("tyrant", ["profession"]),
        ("pharaoh", ["profession"]),
        ("sultan", ["profession"]),
        ("shogun", ["profession"]),
        ("samurai", ["profession"]),
        ("ninja", ["profession"]),
        ("pirate", ["profession"]),
        ("viking", ["profession"]),
        ("knight", ["profession"]),
        ("wizard", ["profession"]),
        ("witch", ["profession"]),
        ("sorcerer", ["profession"]),
        ("warlock", ["profession"]),
        ("cleric", ["profession"]),
        ("monk", ["profession"]),
        ("bard", ["profession"]),
        ("druid", ["profession"]),
        ("ranger", ["profession"]),
        ("paladin", ["profession"]),
        ("barbarian", ["profession"]),
        ("fighter", ["profession"]),
        ("rogue", ["profession"]),
        ("thief", ["profession"]),
        ("assassin", ["profession"]),
        ("necromancer", ["profession"]),
        ("alchemist", ["profession"]),
        ("blacksmith", ["profession"]),
        ("merchant", ["profession"]),
        ("farmer", ["profession"]),
        ("hunter", ["profession"]),
        ("fisherman", ["profession"]),
        ("lumberjack", ["profession"]),
        ("shoe", ["thing"]),
        ("sock", ["thing"]),
        ("shirt", ["thing"]),
        ("pants", ["thing"]),
        ("shorts", ["thing"]),
        ("skirt", ["thing"]),
        ("dress", ["thing"]),
        ("suit", ["thing"]),
        ("tie", ["thing"]),
        ("bowtie", ["thing"]),
        ("scarf", ["thing"]),
        ("gloves", ["thing"]),
        ("mittens", ["thing"]),
        ("hat", ["thing"]),
        ("cap", ["thing"]),
        ("helmet", ["thing"]),
        ("glasses", ["thing"]),
        ("sunglasses", ["thing"]),
        ("watch", ["thing"]),
        ("bracelet", ["thing"]),
        ("necklace", ["thing"]),
        ("earrings", ["thing"]),
        ("ring", ["thing"]),
        ("belt", ["thing"]),
        ("wallet", ["thing"]),
        ("purse", ["thing"]),
        ("backpack", ["thing"]),
        ("bag", ["thing"]),
        ("suitcase", ["thing"]),
        ("briefcase", ["thing"]),
        ("book", ["thing"]),
        ("notebook", ["thing"]),
        ("journal", ["thing"]),
        ("diary", ["thing"]),
        ("calendar", ["thing"]),
        ("planner", ["thing"]),
        ("pen", ["thing"]),
        ("pencil", ["thing"]),
        ("marker", ["thing"]),
        ("highlighter", ["thing"]),
        ("eraser", ["thing"]),
        ("ruler", ["thing"]),
        ("scissors", ["thing"]),
        ("tape", ["thing"]),
        ("glue", ["thing"]),
        ("stapler", ["thing"]),
        ("paper", ["thing"]),
        ("envelope", ["thing"]),
        ("folder", ["thing"]),
        ("file", ["thing"]),
        ("computer", ["thing"]),
        ("laptop", ["thing"]),
        ("tablet", ["thing"]),
        ("phone", ["thing"]),
        ("smartphone", ["thing"]),
        ("smartwatch", ["thing"]),
        ("headphones", ["thing"]),
        ("earbuds", ["thing"]),
        ("speaker", ["thing"]),
        ("microphone", ["thing"]),
        ("camera", ["thing"]),
        ("television", ["thing"]),
        ("monitor", ["thing"]),
        ("keyboard", ["thing"]),
        ("mouse", ["thing"]),
        ("trackpad", ["thing"]),
    ],
    "adjectives": [
        ("blue", ["color"]),
        ("green", ["color"]),
        ("purple", ["color"]),
        ("orange", ["color"]),
        ("black", ["color"]),
        ("white", ["color"]),
        ("gray", ["color"]),
        ("brown", ["color"]),
        ("pink", ["color"]),
        ("tan", ["color"]),
        ("cyan", ["color"]),
        ("magenta", ["color"]),
        ("olive", ["color"]),
        ("maroon", ["color"]),
        ("navy", ["color"]),
        ("aquamarine", ["color"]),
        ("turquoise", ["color"]),
        ("silver", ["color"]),
        ("lime", ["color"]),
        ("teal", ["color"]),
        ("indigo", ["color"]),
        ("violet", ["color"]),
        ("red", ["color"]),
        ("yellow", ["color"]),
        ("abundant", ["quantity"]),
        ("ample", ["quantity"]),
        ("bountiful", ["quantity"]),
        ("copious", ["quantity"]),
        ("countless", ["quantity"]),
        ("endless", ["quantity"]),
        ("enough", ["quantity"]),
        ("excessive", ["quantity"]),
        ("few", ["quantity"]),
        ("infinite", ["quantity"]),
        ("many", ["quantity"]),
        ("numerous", ["quantity"]),
        ("plentiful", ["quantity"]),
        ("sufficient", ["quantity"]),
        ("surplus", ["quantity"]),
        ("unlimited", ["quantity"]),
        ("massive", ["size"]),
        ("tiny", ["size"]),
        ("small", ["size"]),
        ("medium", ["size"]),
        ("large", ["size"]),
        ("huge", ["size"]),
        ("gigantic", ["size"]),
        ("enormous", ["size"]),
        ("minuscule", ["size"]),
        ("microscopic", ["size"]),
        ("massive", ["size"]),
        ("giant", ["size"]),
        ("titanic", ["size"]),
        ("colossal", ["size"]),
        ("immense", ["size"]),
        ("monstrous", ["size"]),
        ("vast", ["size"]),
        ("tremendous", ["size"]),
        ("angry", ["emotion"]),
        ("happy", ["emotion"]),
        ("sad", ["emotion"]),
        ("excited", ["emotion"]),
        ("bored", ["emotion"]),
        ("scared", ["emotion"]),
        ("surprised", ["emotion"]),
        ("confused", ["emotion"]),
        ("disgusted", ["emotion"]),
        ("embarrassed", ["emotion"]),
        ("envious", ["emotion"]),
        ("frightened", ["emotion"]),
        ("guilty", ["emotion"]),
        ("jealous", ["emotion"]),
        ("lonely", ["emotion"]),
        ("proud", ["emotion"]),
        ("shy", ["emotion"]),
        ("tired", ["emotion"]),
        ("worried", ["emotion"]),
        ("anxious", ["emotion"]),
        ("ashamed", ["emotion"]),
        ("bored", ["emotion"]),
        ("calm", ["emotion"]),
        ("content", ["emotion"]),
        ("delighted", ["emotion"]),
        ("ecstatic", ["emotion"]),
        ("elated", ["emotion"]),
        ("euphoric", ["emotion"]),
        ("exhilarated", ["emotion"]),
        ("glad", ["emotion"]),
        ("joyful", ["emotion"]),
        ("jubilant", ["emotion"]),
        ("lively", ["emotion"]),
        ("merry", ["emotion"]),
        ("overjoyed", ["emotion"]),
        ("peaceful", ["emotion"]),
        ("pleased", ["emotion"]),
        ("radiant", ["emotion"]),
        ("satisfied", ["emotion"]),
        ("serene", ["emotion"]),
        ("thrilled", ["emotion"]),
        ("upbeat", ["emotion"]),
        ("vibrant", ["emotion"]),
        ("zestful", ["emotion"]),
        ("zesty", ["emotion"]),
        ("zippy", ["emotion"]),
        ("broken", ["condition"]),
        ("clean", ["condition"]),
        ("dirty", ["condition"]),
        ("empty", ["condition"]),
        ("full", ["condition"]),
        ("hot", ["condition"]),
        ("cold", ["condition"]),
        ("wet", ["condition"]),
        ("dry", ["condition"]),
        ("fast", ["condition"]),
        ("slow", ["condition"]),
        ("hard", ["condition"]),
        ("soft", ["condition"]),
        ("heavy", ["condition"]),
        ("light", ["condition"]),
        ("noisy", ["condition"]),
        ("quiet", ["condition"]),
        ("bright", ["condition"]),
        ("dark", ["condition"]),
        ("thick", ["condition"]),
        ("thin", ["condition"]),
        ("sharp", ["condition"]),
        ("dull", ["condition"]),
        ("smooth", ["condition"]),
        ("rough", ["condition"]),
        ("sweet", ["condition"]),
        ("sour", ["condition"]),
        ("bitter", ["condition"]),
        ("salty", ["condition"]),
        ("spicy", ["condition"]),
        ("savory", ["condition"]),
        ("tangy", ["condition"]),
        ("tart", ["condition"]),
        ("mild", ["condition"]),
        ("hot", ["condition"]),
        ("cold", ["condition"]),
        ("warm", ["condition"]),
        ("cool", ["condition"]),
        ("icy", ["condition"]),
        ("frozen", ["condition"]),
        ("boiling", ["condition"]),
        ("scalding", ["condition"]),
        ("steaming", ["condition"]),
        ("roasting", ["condition"]),
        ("toasty", ["condition"]),
        ("chilly", ["condition"]),
        ("freezing", ["condition"]),
        ("frosty", ["condition"]),
        ("frigid", ["condition"]),
        ("arctic", ["condition"]),
        ("polar", ["condition"]),
        ("tropical", ["condition"]),
        ("subtropical", ["condition"]),
        ("temperate", ["condition"]),
        ("mild", ["condition"]),
        ("moderate", ["condition"]),
        ("extreme", ["condition"]),
        ("severe", ["condition"]),
        ("intense", ["condition"]),
        ("harsh", ["condition"]),
        ("brutal", ["condition"]),
        ("cruel", ["condition"]),
        ("merciless", ["condition"]),
        ("unrelenting", ["condition"]),
        ("unforgiving", ["condition"]),
    ],
}

vocabulary = Vocabulary(**vocabulary_kwargs)


def get_words_by_category(
    part_of_speech: str, categories: Optional[List[Category]] = None
) -> List[WordEntry]:
    words = []
    for word, word_categories in vocabulary[part_of_speech]:
        if categories:
            # check if any of the categories are in word_categories
            if any(category in word_categories for category in categories):
                words.append(word)
        else:
            words.append(word)

    return words



import unittest

from random_slugs.generate import (
    RandomSlugConfigError,
    format_slug,
    generate_slug,
    get_total_unique_slugs,
)
from random_slugs.words import vocabulary


def check_word_in_category(part_of_speech, word, categories) -> bool:
    cats = set(categories)
    for vocab in vocabulary[part_of_speech]:
        if vocab[0] == word:
            return len(cats.intersection(vocab[1])) > 0
    return False


class RandomSlugsTest(unittest.TestCase):
    def test_generate_slug(self):
        slug = generate_slug()
        self.assertIsInstance(slug, str)
        self.assertTrue(len(slug) > 10)

    def test_parts_are_unique(self):
        slug = generate_slug()
        words = slug.split("-")
        self.assertEqual(len(words), len(set(words)))

    def test_number_of_words(self):
        sizes = [2, 3, 4, 5, 6]
        for size in sizes:
            slug = generate_slug(size)
            words = slug.split("-")
            self.assertEqual(len(words), size)

    def test_words_in_specified_category(self):
        options = {
            "categories": {
                "nouns": ["pokemon"],
                "adjectives": ["color"],
            }
        }
        slug = generate_slug(3, options)
        chunks = slug.split("-")
        self.assertTrue(
            check_word_in_category("adjectives", chunks[0], options["categories"]["adjectives"])
        )
        self.assertTrue(
            check_word_in_category("adjectives", chunks[1], options["categories"]["adjectives"])
        )
        self.assertTrue(check_word_in_category("nouns", chunks[2], options["categories"]["nouns"]))

    def test_if_random_seed_is_set_slug_should_be_equal(self):
        options = {"seed": 42}
        slug1 = generate_slug(options=options)
        slug2 = generate_slug(options=options)
        slug3 = generate_slug(options={"seed": 43})
        self.assertEqual(slug1, slug2)
        self.assertNotEqual(slug1, slug3)

    def test_if_random_seed_is_set_slug_should_be_different(self):
        slug1 = generate_slug()
        slug2 = generate_slug()
        self.assertNotEqual(slug1, slug2)

    def test_invalid_options_seed(self):
        with self.assertRaises(RandomSlugConfigError):
            generate_slug(options={"seed": None})

    def test_invalid_options_parts_of_speech(self):
        with self.assertRaises(RandomSlugConfigError):
            generate_slug(options={"parts_of_speech": ["invalid"]})

    def test_invalid_options_categories(self):
        with self.assertRaises(RandomSlugConfigError):
            generate_slug(options={"categories": {"invalid": []}})

    def test_invalid_options_format(self):
        with self.assertRaises(RandomSlugConfigError):
            generate_slug(options={"format": "invalid"})

    def test_invalid_options_seed_type(self):
        with self.assertRaises(RandomSlugConfigError):
            generate_slug(options={"seed": {}})

    def test_empty_options_parts_of_speech_type(self):
        generate_slug(options={"parts_of_speech": {}})


class FormatSlugTest(unittest.TestCase):
    def test_kebab_case(self):
        words = ["red", "panda"]
        slug = format_slug(words, "kebab")
        self.assertEqual(slug, "red-panda")

    def test_snake_case(self):
        words = ["red", "panda"]
        slug = format_slug(words, "snake")
        self.assertEqual(slug, "red_panda")

    def test_camel_case(self):
        words = ["red", "panda"]
        slug = format_slug(words, "camel")
        self.assertEqual(slug, "redPanda")

    def test_default_format(self):
        words = ["red", "panda"]
        slug = format_slug(words)
        self.assertEqual(slug, "red-panda")


class TotalUniqueSlugsTest(unittest.TestCase):
    def setUp(self):
        self.all_adjectives = [v[0] for v in vocabulary["adjectives"]]
        self.all_nouns = [v[0] for v in vocabulary["nouns"]]
        self.num_adjectives = len(self.all_adjectives)
        self.num_nouns = len(self.all_nouns)

    def test_total_unique_slugs(self):
        num = get_total_unique_slugs()
        total = self.num_adjectives * self.num_adjectives * self.num_nouns
        self.assertEqual(num, total)

    def test_total_in_subsets_of_categories(self):
        options = {
            "categories": {
                "nouns": ["animal", "technology"],
                "adjectives": ["color", "size"],
            }
        }
        num = get_total_unique_slugs(4, options)
        num_adjectives = len(
            [
                word
                for word, categories in vocabulary["adjectives"]
                if any(category in categories for category in options["categories"]["adjectives"])
            ]
        )
        num_nouns = len(
            [
                word
                for word, categories in vocabulary["nouns"]
                if any(category in categories for category in options["categories"]["nouns"])
            ]
        )
        total = num_adjectives**3 * num_nouns
        self.assertEqual(num, total)
